# -*- coding: utf-8 -*-

from django.conf import settings
from django.urls import path

from . import views

urlpatterns = [
    # sync urls
    path(settings.DJANKISERV_SYNC_URLBASE + "start", views.base_start, name="sync_start"),
    path(settings.DJANKISERV_SYNC_URLBASE + "applyChanges", views.base_applyChanges, name="sync_applyChanges"),
    path(settings.DJANKISERV_SYNC_URLBASE + "applyGraves", views.base_applyGraves, name="sync_applyGraves"),
    path(settings.DJANKISERV_SYNC_URLBASE + "chunk", views.base_chunk, name="sync_chunk"),
    path(settings.DJANKISERV_SYNC_URLBASE + "applyChunk", views.base_applyChunk, name="sync_applyChunk"),
    path(settings.DJANKISERV_SYNC_URLBASE + "sanityCheck2", views.base_sanityCheck2, name="sync_sanityCheck2"),
    path(settings.DJANKISERV_SYNC_URLBASE + "finish", views.base_finish, name="sync_finish"),
    path(settings.DJANKISERV_SYNC_URLBASE + "meta", views.base_meta, name="meta"),
    path(settings.DJANKISERV_SYNC_URLBASE + "hostKey", views.base_hostKey, name="hostKey"),
    path(settings.DJANKISERV_SYNC_URLBASE + "upload", views.base_upload, name="upload"),
    path(settings.DJANKISERV_SYNC_URLBASE + "download", views.base_download, name="download"),
    path(settings.DJANKISERV_SYNC_MEDIA_URLBASE + "begin", views.media_begin, name="media_begin"),
    path(settings.DJANKISERV_SYNC_MEDIA_URLBASE + "mediaChanges", views.media_mediaChanges, name="media_mediaChanges"),
    path(settings.DJANKISERV_SYNC_MEDIA_URLBASE + "mediaSanity", views.media_mediaSanity, name="media_mediaSanity"),
    path(
        settings.DJANKISERV_SYNC_MEDIA_URLBASE + "uploadChanges", views.media_uploadChanges, name="media_uploadChanges"
    ),
    path(
        settings.DJANKISERV_SYNC_MEDIA_URLBASE + "downloadFiles", views.media_downloadFiles, name="media_downloadFiles"
    ),
]



# -*- coding: utf-8 -*-

from django.apps import AppConfig


class DjankiservConfig(AppConfig):
    name = "djankiserv"

    def ready(self):
        import djankiserv.signals  # noqa # pylint: disable=C0415,W0611



# -*- coding: utf-8 -*-

from django.conf import settings
from django.urls import path
from rest_framework_simplejwt.views import TokenObtainPairView, TokenRefreshView

from . import views

urlpatterns = [
    # JWT
    path(settings.DJANKISERV_API_URLBASE + "token", TokenObtainPairView.as_view(), name="token_obtain_pair"),
    path(settings.DJANKISERV_API_URLBASE + "token/refresh", TokenRefreshView.as_view(), name="token_refresh"),
    # api urls
    path(settings.DJANKISERV_API_URLBASE + "addNotes", views.add_notes, name="add_notes"),
    path(settings.DJANKISERV_API_URLBASE + "deleteNotes", views.delete_notes, name="delete_notes"),
    path(settings.DJANKISERV_API_URLBASE + "decks", views.decks, name="decks"),
    path(settings.DJANKISERV_API_URLBASE + "decksConf", views.decks_conf, name="decks_conf"),
    path(settings.DJANKISERV_API_URLBASE + "tags", views.tags, name="tags"),
    path(settings.DJANKISERV_API_URLBASE + "models", views.models, name="models"),
    path(settings.DJANKISERV_API_URLBASE + "notes", views.notes, name="notes"),
]



# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import djankiserv.unki

from . import intTime

# Cards
##########################################################################

# Type: 0=new, 1=learning, 2=due
# Queue: same as above, and:
#        -1=suspended, -2=user buried, -3=sched buried
# Due is used differently for different queues.
# - new queue: note id or random int
# - rev queue: integer day
# - lrn queue: integer timestamp


class Card:  # pylint: disable=R0902
    def __init__(self, col, card_id=None):
        self.col = col
        self.timerStarted = None
        self._qa = None
        self._note = None
        self.mod = None
        self.usn = None

        if card_id:
            self.id = card_id
            self.load()
        else:
            # to flush, set nid, ord, and due
            self.id = col.timestamp_for_table("cards")
            self.did = 1
            self.crt = intTime()
            self.type = 0
            self.queue = 0
            self.ivl = 0
            self.factor = 0
            self.reps = 0
            self.lapses = 0
            self.left = 0
            self.odue = 0
            self.odid = 0
            self.flags = 0
            self.data = ""

    def flush(self):
        self.mod = intTime()
        self.usn = self.col.usn
        # bug check
        if self.queue == 2 and self.odue and not self.col.decks.isDyn(self.did):
            raise Exception("odueInvalid")  # FIXME: here upstream ran runHook("odueInvalid")
        assert self.due < 4294967296

        self.col.db.execute(
            djankiserv.unki.AnkiDataModel.insert_on_conflict_update(self.col.username, "cards"),
            self.id,
            self.nid,
            self.did,
            self.ord,
            self.mod,
            self.usn,
            self.type,
            self.queue,
            self.due,
            self.ivl,
            self.factor,
            self.reps,
            self.lapses,
            self.left,
            self.odue,
            self.odid,
            self.flags,
            self.data,
        )

    def load(self):
        (
            self.id,
            self.nid,
            self.did,
            self.ord,
            self.mod,
            self.usn,
            self.type,
            self.queue,
            self.due,
            self.ivl,
            self.factor,
            self.reps,
            self.lapses,
            self.left,
            self.odue,
            self.odid,
            self.flags,
            self.data,
        ) = self.col.db.first(f"select * from {self.col.username}.cards where id = %s", self.id)
        self._qa = None
        self._note = None



# -*- coding: utf-8 -*-

import json
import logging
import os
import pathlib
import pickle
import subprocess

from django.conf import settings
from django.db import connections
from django.http import HttpResponse

import djankiserv.unki
from djankiserv.assets import jsonfiles  # noqa: 401  # pylint: disable=W0611

from . import AnkiDataModelBase, get_data

MODEL_STD = 0
NEW_CARDS_DUE = 1

logger = logging.getLogger("djankiserv.unki.database")


def db_conn():
    return connections["userdata"]


class StandardDB:
    def __init__(self):  # pylint: disable=W0231
        self._db = db_conn()
        self.mod = False

    @staticmethod
    def schema_exists(schema_name):
        with db_conn().cursor() as cur:
            cur.execute("SELECT 1 FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = %s", (schema_name,))
            res = cur.fetchone()
        return res[0] if res else 0

    @staticmethod
    def create_schema(schema_name):
        assert schema_name.isidentifier()
        with db_conn().cursor() as cur:
            for sql in djankiserv.unki.AnkiDataModel.generate_schema_sql_list(schema_name):
                if not sql.strip():
                    continue
                # print("executing", sql)
                cur.execute(sql)
            res = cur.fetchone()
            return res[0]  # returns an Ok message
        # db_conn().commit()

    @staticmethod
    def delete_schema(schema_name):
        with db_conn().cursor() as cur:
            cur.execute(djankiserv.unki.AnkiDataModel.DROP_SCHEMA.replace("{schema_name}", schema_name))

    def execute(self, sql, *a, **ka):
        s = sql.strip().lower()
        # mark modified?
        for stmt in "insert", "update", "delete":
            if s.startswith(stmt):
                self.mod = True
        cur = self._db.cursor()
        if ka:
            cur.execute(sql, ka)
        else:
            cur.execute(sql, a)
        return cur

    def executemany(self, sql, li):
        self.mod = True
        cur = self._db.cursor()
        cur.executemany(sql, li)
        return cur

    def commit(self):
        pass
        # self._db.commit()

    def scalar(self, *a):
        res = self.execute(*a).fetchone()  # pylint: disable=E1120  # FIXME: how should I get rid of this disable?
        if res:
            return res[0]
        return None

    def first(self, *a):
        c = self.execute(*a)
        res = c.fetchone()
        c.close()
        return res

    def list(self, *a):
        return [x[0] for x in self.execute(*a)]  # pylint: disable=E1120  # FIXME: how should I get rid of this disable?

    def close(self):
        # FIXME: is this Ok?
        pass
        # self._db.commit()
        # self._db.close()


# for getting dumps of the DB to compare for unittests
def dump_io_to_file(session, method, io_obj, is_media=False):  # noqa: C901  # pylint: disable=R0914
    if not settings.DJANKISERV_GENERATE_TEST_ASSETS:  # we need this only for getting new test files
        return
    schema_name = session["name"]
    fname_dir = session["dump_base"] if not is_media else session["dump_base_media"]
    pathlib.Path(fname_dir).mkdir(parents=True, exist_ok=True)

    if isinstance(io_obj, HttpResponse):
        fname_base = os.path.join(fname_dir, "post_" + method)
        to_print = f"Response Content:\n{io_obj.content}\n"
        try:
            io_obj_json = json.loads(io_obj.content)
        except (json.decoder.JSONDecodeError, UnicodeDecodeError):
            io_obj_json = {"to_return": io_obj.content}  # this will get pickled below
    else:
        fname_base = os.path.join(fname_dir, "pre_" + method)
        to_print = f"POST:\n{io_obj.POST}\n\n"
        to_print += f"DATA:\n{get_data(io_obj)}\n"
        io_obj_json = get_data(io_obj)

    if os.path.exists(fname_base + ".txt"):
        i = 1
        while os.path.exists(f"{fname_base}.{i}.txt"):
            i += 1
        fname_base = f"{fname_base}.{i}"

    with open(fname_base + ".txt", "w") as fh:
        fh.write(to_print)

    proc_res = subprocess.run(
        ["pg_dump", connections["userdata"].settings_dict["NAME"], "--schema", schema_name, "--column-inserts"],
        capture_output=True,
        text=True,
        check=True,
    )

    with open(fname_base + ".sql", "w") as fh:
        fh.write(proc_res.stdout)

    try:
        with open(fname_base + ".json", "w") as fh:
            json.dump(io_obj_json, fh, indent=4, sort_keys=True)
    except TypeError:
        # there is binary data in the json, so pickle instead
        with open(fname_base + ".bin", "wb") as fh:
            pickle.dump(io_obj_json, fh)
        try:  # to remove the broken .json file that was created
            os.remove(fname_base + ".json")
        except OSError:
            pass

    with open(fname_base + "_col.json", "w") as fh, db_conn().cursor() as cur:
        cur.execute(f"select conf, models, decks, dconf, tags from {schema_name}.col")
        row = cur.fetchall()
        if len(row) > 1:
            raise Exception("There should only ever be one collection, right?")
        conf = json.loads(row[0][0])
        models = json.loads(row[0][1])
        decks = json.loads(row[0][2])
        dconf = json.loads(row[0][3])
        tags = json.loads(row[0][4])
        fh.write(
            json.dumps(
                {"conf": conf, "models": models, "decks": decks, "dconf": dconf, "tags": tags},
                sort_keys=True,
                indent=4,
            )
        )


class PostgresAnkiDataModel(AnkiDataModelBase):
    MODEL_SETUP = """
        SET statement_timeout = 0;
        SET lock_timeout = 0;
        SET idle_in_transaction_session_timeout = 0;
        SET client_encoding = 'UTF8';
        SET standard_conforming_strings = on;
        SELECT pg_catalog.set_config('search_path', '', false);
        SET check_function_bodies = false;
        SET client_min_messages = warning;
        SET row_security = off;

        SET default_tablespace = '';

        SET default_with_oids = false;"""

    MODEL_SETUP_LIST = MODEL_SETUP.split(";")
    CREATE_SCHEMA = "CREATE SCHEMA {schema_name};"
    DROP_SCHEMA = "DROP SCHEMA IF EXISTS {schema_name} CASCADE;"

    @staticmethod
    def generate_schema_sql_list(schema_name):
        sql = []
        sql += PostgresAnkiDataModel.MODEL_SETUP_LIST
        sql.append(PostgresAnkiDataModel.CREATE_SCHEMA.format(schema_name=schema_name))

        for table_name, defin in AnkiDataModelBase.MODEL.items():
            tsql = (
                f"CREATE TABLE {schema_name}.{table_name} ("
                + ",".join(
                    [f"{f['name']} {f['type']} {'' if f.get('nullable') else 'NOT'} NULL" for f in defin["fields"]]
                )
                + ");"
            )
            sql.append(tsql)
            identity = [x["name"] for x in defin["fields"] if "is_pk" in x and x["type"] == "bigint"]
            if identity:
                tsql = f"""
                ALTER TABLE {schema_name}.{table_name}
                    ALTER COLUMN {identity[0]} ADD GENERATED BY DEFAULT AS IDENTITY (
                        SEQUENCE NAME {schema_name}.{table_name}_{identity[0]}_seq
                        START WITH 1
                        INCREMENT BY 1
                        NO MINVALUE
                        NO MAXVALUE
                        CACHE 1
                    );"""

                sql.append(tsql)

            pk = [x["name"] for x in defin["fields"] if "is_pk" in x]
            if pk:
                tsql = f"""
                ALTER TABLE ONLY {schema_name}.{table_name}
                    ADD CONSTRAINT {table_name}_pkey PRIMARY KEY ({pk[0]});"""
                sql.append(tsql)

            for i in defin["indexes"]:
                tsql = (
                    f"CREATE INDEX {i['name']} ON {schema_name}.{table_name} ("  # defaults to btree for postgresql
                    + ",".join([x["name"] for x in i["fields"]])
                    + ");"
                )
                sql.append(tsql)

            if "initsql" in defin:
                tsql = defin["initsql"].replace("{schema_name}", schema_name)  # dynamic replace rather than fstring
                sql.append(tsql)

        sql.append(f"SELECT {AnkiDataModelBase.VERSION};")

        return sql

    @staticmethod
    def insert_on_conflict_update(schema_name, table_name):
        identity = [x["name"] for x in AnkiDataModelBase.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(AnkiDataModelBase.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in AnkiDataModelBase.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON CONFLICT ({identity[0]}) DO UPDATE SET "
            + ",".join(
                [
                    f"{f['name']} = EXCLUDED.{f['name']}"
                    for f in AnkiDataModelBase.MODEL[table_name]["fields"]
                    if "is_pk" not in f
                ]
            )
        )

    @staticmethod
    def insert_on_conflict_nothing(schema_name, table_name):
        identity = [x["name"] for x in AnkiDataModelBase.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(AnkiDataModelBase.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in AnkiDataModelBase.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON CONFLICT ({identity[0]}) DO NOTHING "
        )

    @staticmethod
    def replace_schema(cur, to_replace_name, replace_with_name):
        # rename the existing schema, rename the new schema to the username, delete the old schema
        cur.execute(f"ALTER SCHEMA {to_replace_name} RENAME TO {to_replace_name}_old")
        cur.execute(f"ALTER SCHEMA {replace_with_name} RENAME TO {to_replace_name}")
        cur.execute(f"DROP SCHEMA {to_replace_name}_old CASCADE")


class MariadbAnkiDataModel(AnkiDataModelBase):

    MODEL_SETUP = ""
    MODEL_SETUP_LIST = []

    CREATE_SCHEMA = "CREATE DATABASE {schema_name} CHARACTER SET utf8;"
    DROP_SCHEMA = "DROP DATABASE IF EXISTS {schema_name};"

    @staticmethod
    def generate_schema_sql_list(schema_name):
        sql = []
        sql += MariadbAnkiDataModel.MODEL_SETUP_LIST
        sql.append(MariadbAnkiDataModel.CREATE_SCHEMA.format(schema_name=schema_name))

        for table_name, defin in AnkiDataModelBase.MODEL.items():
            tsql = (
                f"CREATE TABLE {schema_name}.{table_name} ("
                + ",".join(
                    [f"{f['name']} {f['type']} {'' if f.get('nullable') else 'NOT'} NULL" for f in defin["fields"]]
                )
                + ");"
            )
            sql.append(tsql)
            identity = [x["name"] for x in defin["fields"] if "is_pk" in x and x["type"] == "bigint"]
            if identity:
                tsql = (
                    f"""ALTER TABLE {schema_name}.{table_name} MODIFY """
                    f"""{identity[0]} BIGINT NOT NULL AUTO_INCREMENT PRIMARY KEY"""
                )

                sql.append(tsql)
            else:
                pk = [x for x in defin["fields"] if "is_pk" in x]
                if pk:
                    pk_colspec = pk[0]["name"] + "(255)" if pk[0]["type"] == "text" else ""
                    tsql = f"""
                    ALTER TABLE {schema_name}.{table_name}
                        ADD CONSTRAINT {table_name}_pkey PRIMARY KEY ({pk_colspec});"""
                    sql.append(tsql)

            for i in defin["indexes"]:
                tsql = (
                    f"CREATE INDEX {i['name']} ON {schema_name}.{table_name} ("  # defaults to btree for postgresql
                    + ",".join([x["name"] + ("(255)" if x["type"] == "text" else "") for x in i["fields"]])
                    + ");"
                )
                sql.append(tsql)

            if "initsql" in defin:
                tsql = defin["initsql"].replace("{schema_name}", schema_name)  # dynamic replace rather than fstring
                sql.append(tsql)

            # sql += tsql
        sql.append(f"SELECT {AnkiDataModelBase.VERSION};")

        return sql

    @staticmethod
    def insert_on_conflict_update(schema_name, table_name):

        fstr = ", ".join(["%s"] * len(MariadbAnkiDataModel.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + "ON DUPLICATE KEY UPDATE "
            + ",".join(
                [
                    f"{f['name']} = VALUES({f['name']})"
                    for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]
                    if "is_pk" not in f
                ]
            )
        )

    @staticmethod
    def insert_on_conflict_nothing(schema_name, table_name):

        identity = [x["name"] for x in MariadbAnkiDataModel.MODEL[table_name]["fields"] if "is_pk" in x]
        fstr = ", ".join(["%s"] * len(MariadbAnkiDataModel.MODEL[table_name]["fields"]))

        return (
            f"INSERT INTO {schema_name}.{table_name} ("
            + ",".join([f"{f['name']}" for f in MariadbAnkiDataModel.MODEL[table_name]["fields"]])
            + f") VALUES ({fstr}) "
            + f"ON DUPLICATE KEY UPDATE {identity[0]} = {identity[0]}"
        )

    @staticmethod
    def replace_schema(cur, to_replace_name, replace_with_name):
        # rename the existing schema, rename the new schema to the username, delete the old schema

        sql = f"""
            SELECT CONCAT('RENAME TABLE ',
                table_schema, '.', table_name, ' TO ', '{to_replace_name}.', table_name, ';') as commands
            FROM information_schema.TABLES
            WHERE table_schema = '{replace_with_name}';
            """
        cur.execute(sql)
        copy_commands = cur.fetchall()

        sql = f"DROP DATABASE {to_replace_name}; CREATE DATABASE {to_replace_name} CHARACTER SET utf8;"
        cur.execute(sql)

        for com in copy_commands:
            cur.execute(com[0])

        sql = f"DROP DATABASE {replace_with_name};"
        cur.execute(sql)



# -*- coding: utf-8 -*-

from django.conf import settings
from django.contrib.auth.models import User
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from .unki.collection import Collection


@receiver(post_save, sender=User)  # pylint: disable=W0613
def update_profile_signal(sender, instance, created, **kwargs):  # pylint: disable=W0613
    if created:
        # Creates on first loading if it doesn't exist
        # FIXME: so do I still need to do this here?
        Collection(instance.username, settings.DJANKISERV_DATA_ROOT)


@receiver(post_delete, sender=User)
def delete_user_signal(sender, instance, **kwargs):  # pylint: disable=W0613
    Collection.delete(instance.username, settings.DJANKISERV_DATA_ROOT)  # pylint: disable=W0613



# -*- coding: utf-8 -*-

import copy
import datetime
import json
import logging
import os
import pathlib
import pkgutil
import random
import re
import shutil
import time
import unicodedata

import djankiserv.unki
from djankiserv.assets import jsonfiles  # noqa: 401  # pylint: disable=W0611
from djankiserv.unki.database import StandardDB

from . import REM_CARD, REM_NOTE, checksum, fieldChecksum, ids2str, intTime, joinFields, splitFields, stripHTMLMedia
from .cards import Card
from .decks import DeckManager
from .models import ModelManager
from .notes import Note
from .sched import Scheduler

MODEL_STD = 0
NEW_CARDS_DUE = 1

logger = logging.getLogger("djankiserv.unki.collection")


class Collection:  # pylint: disable=R0902,R0904
    def __init__(self, username, media_dir_base):
        self.username = username
        self._media_dir = os.path.join(media_dir_base, self.username)
        self.tags = None  # to make pylint happy
        self.tags_changed = False  # to make pylint happy
        self.mod = 0  # to make pylint happy

        self.db = StandardDB()
        self.models = ModelManager(self)
        self.decks = DeckManager(self)

        create = not StandardDB.schema_exists(self.username)
        if create:
            StandardDB.create_schema(self.username)
            self.db.execute(f"update {self.username}.col set scm = %s", intTime(1000))

            c = json.loads(
                pkgutil.get_data("djankiserv.assets.jsonfiles", "default_collection_conf.json").decode("utf-8")
            )
            g = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_deck.json").decode("utf-8"))
            # g["mod"] = intTime() FIXME: is this necessary??? can we leave the original mod times?
            gc = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_deck_conf.json").decode("utf-8"))

            model = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_model.json").decode("utf-8"))

            self.db.execute(
                f"update {self.username}.col set conf = %s, decks = %s, dconf = %s, models = %s",
                json.dumps(c),
                json.dumps({"1": g}),
                json.dumps({"1": gc}),
                json.dumps(model),
            )

        self._lastSave = time.time()
        self.load()

        if not self.crt:
            d = datetime.datetime.today()
            d -= datetime.timedelta(hours=4)
            d = datetime.datetime(d.year, d.month, d.day)
            d += datetime.timedelta(hours=4)
            self.crt = int(time.mktime(d.timetuple()))

        self.save()

        self.sched = Scheduler(self)

        if not self.conf.get("newBury", False):
            self.conf["newBury"] = True
            self.db.mod = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, *args):
        # FIXME: should this be self.close(save=True)?
        self.close()

    @staticmethod
    def delete(username, media_dir_base):
        shutil.rmtree(os.path.join(media_dir_base, username), ignore_errors=True)
        StandardDB.delete_schema(username)

    def media_changes(self, client_last_usn):

        result = []
        server_last_usn = self.last_media_usn()

        if client_last_usn < server_last_usn or client_last_usn == 0:
            for fname, usn, csum, in self.db.execute(
                f"select fname,usn,csum from {self.username}.media order by usn desc limit %s",
                server_last_usn - client_last_usn,
            ):
                result.append([fname, usn, csum])

        result.reverse()
        return result

    def media_dir(self):
        return self._media_dir

    def last_media_usn(self):
        return self.db.scalar(f"SELECT max(usn) FROM {self.username}.media") or 0

    def media_count(self):
        return self.db.scalar(f"SELECT count(0) FROM {self.username}.media WHERE csum IS NOT NULL")

    def media_sync_delete(self, fname):
        fpath = os.path.join(self.media_dir(), fname)
        if os.path.exists(fpath):
            os.remove(fpath)
        self.db.execute(
            f"UPDATE {self.username}.media SET csum = NULL, usn = %s WHERE fname = %s",
            self.last_media_usn() + 1,
            fname,
        )

    @staticmethod
    def check_zip_data(zip_file):
        max_zip_size = 100 * 1024 * 1024
        max_meta_file_size = 100000

        meta_file_size = zip_file.getinfo("_meta").file_size
        sum_file_sizes = sum(info.file_size for info in zip_file.infolist())

        if meta_file_size > max_meta_file_size:
            raise ValueError("Zip file's metadata file is larger than %s " "Bytes." % max_meta_file_size)
        if sum_file_sizes > max_zip_size:
            raise ValueError("Zip file contents are larger than %s Bytes." % max_zip_size)

    def _remove_media_files(self, filenames):
        """
        Marks all files in list filenames as deleted and removes them from the
        media directory.
        """
        logger.debug("Removing %d files from media dir.", len(filenames))
        for filename in filenames:
            try:
                self.media_sync_delete(filename)
                self.db.commit()
            except OSError as err:
                logger.error("Error when removing file '%s' from media dir: " "%s", filename, str(err))

    def adopt_media_changes_from_zip(self, zip_file):  # pylint: disable=R0914
        """
        Adds and removes files to/from the database and media directory
        according to the data in zip file zipData.
        """

        # Get meta info first.
        meta = json.loads(zip_file.read("_meta").decode())

        # Remove media files that were removed on the client.
        media_to_remove = []
        for normname, ordinal in meta:
            if not ordinal:
                media_to_remove.append(unicodedata.normalize("NFC", normname))

        # Add media files that were added on the client.
        media_to_add = []
        usn = self.last_media_usn()
        oldUsn = usn
        for i in zip_file.infolist():
            if i.filename == "_meta":  # Ignore previously retrieved metadata.
                continue

            file_data = zip_file.read(i)
            csum = checksum(file_data)
            filename = unicodedata.normalize("NFC", meta[int(i.filename)][0])

            # FIXME: need to completely redo everything regarding media files!!!
            pathlib.Path(self.media_dir()).mkdir(parents=True, exist_ok=True)
            file_path = os.path.join(self.media_dir(), filename)

            # Save file to media directory.
            with open(file_path, "wb") as f:
                f.write(file_data)

            usn += 1
            media_to_add.append((filename, usn, csum))

        # We count all files we are to remove, even if we don't have them in
        # our media directory and our db doesn't know about them.
        processed_count = len(media_to_remove) + len(media_to_add)

        assert len(meta) == processed_count  # sanity check

        if media_to_remove:
            self._remove_media_files(media_to_remove)

        if media_to_add:
            sql = djankiserv.unki.AnkiDataModel.insert_on_conflict_update(self.username, "media")
            self.db.executemany(sql, media_to_add)
            self.db.commit()

        assert self.last_media_usn() == oldUsn + processed_count  # TODO: move to some unit test
        return processed_count

    def reopen(self):
        "Reconnect to DB"
        if not self.db:
            self.db = StandardDB()

    def load_tags(self, json_):
        self.tags = json.loads(json_)
        self.tags_changed = False

    def flush_tags(self):
        if self.tags_changed:
            self.db.execute(f"update {self.username}.col set tags=%s", json.dumps(self.tags))
            self.tags_changed = False

    def register_tags(self, tags, usn=None):
        "Given a list of tags, add any missing ones to tag registry."
        for t in tags:
            if t not in self.tags:
                self.tags[t] = self.usn if usn is None else usn
                self.tags_changed = True

    def all_tags(self):
        return list(self.tags.items())

    def tagstring_for_note(self, note_tags):
        "Strip duplicates, adjust case to match existing tags, and sort."
        stripped_tags = []
        for t in note_tags:
            s = re.sub("[\"']", "", t)
            for existing_tag in self.tags:
                if s.lower() == existing_tag.lower():
                    s = existing_tag
            stripped_tags.append(s)
        tags = sorted(set(stripped_tags))
        if not tags:
            return ""
        return " %s " % " ".join(tags)

    def load(self):
        (
            self.crt,
            self.mod,
            self.scm,
            self.dty,  # no longer used
            self.usn,
            self.ls,
            self.conf,
            models,
            decks,
            dconf,
            tags,
        ) = self.db.first(
            f"""
            select crt, modified, scm, dty, usn, ls,
            conf, models, decks, dconf, tags from {self.username}.col"""
        )
        self.conf = json.loads(self.conf)

        # FIXME: remove when
        # https://stackoverflow.com/questions/63760777/psycopg2-vs-mysqldb-backslash-escaping-behaviour
        # has an answer. This is due to exec'ing the inserts for the tests NOT escaping the backslashes in
        # mariadb, so when they try to get loaded they get interpreted as escapes in the json...
        # try:
        self.models.load(models)
        # except json.decoder.JSONDecodeError:
        #     self.models.load(repr(models)[1:-1])

        self.decks.load(decks, dconf)
        self.load_tags(tags)

    def flush(self, mod=None):
        "Flush state to DB, updating mod time."
        self.mod = intTime(1000) if mod is None else mod
        self.db.execute(
            f"""update {self.username}.col set
                crt=%s, modified=%s, scm=%s, dty=%s, usn=%s, ls=%s, conf=%s""",
            self.crt,
            self.mod,
            self.scm,
            self.dty,
            self.usn,
            self.ls,
            json.dumps(self.conf),
        )

    def save(self, mod=None):
        "Flush, commit DB, and take out another write lock."
        # let the managers conditionally flush
        self.models.flush()
        self.decks.flush()
        self.flush_tags()
        # and flush deck + bump mod if db has been changed
        if self.db.mod:
            self.flush(mod=mod)
            self.db.commit()
            # self.lock()
            self.db.mod = False
        self._lastSave = time.time()

    def close(self, save=True):
        "Disconnect from DB."
        if self.db:
            if save:
                self.save()

            self.db.close()
            self.db = None

    def add_graves(self, ids, obj_type):
        self.db.executemany(
            f"insert into {self.username}.graves values ({self.usn}, %s, {obj_type})", ([x] for x in ids)
        )

    def get_notes(self, note_ids=None):
        where = ""
        if note_ids:
            where = f" where id in {ids2str(note_ids)}"
        return [Note(self, note_id=i) for i in self.db.list(f"select id from {self.username}.notes {where}")]

    def all_note_ids(self):
        return self.db.list(f"select id from {self.username}.notes")

    def rem_notes(self, ids):
        "Bulk delete notes by ID. Don't call this directly."
        if not ids:
            return
        strids = ids2str(ids)
        # we need to log these independently of cards, as one side may have
        # more card templates
        self.add_graves(ids, REM_NOTE)
        self.db.execute(f"delete from {self.username}.notes where id in {strids}")

    def rem_cards(self, ids, notes=True):
        "Bulk delete cards by ID."
        if not ids:
            return
        sids = ids2str(ids)
        nids = self.db.list(f"select nid from {self.username}.cards where id in " + sids)
        # remove cards
        self.add_graves(ids, REM_CARD)
        self.db.execute(f"delete from {self.username}.cards where id in " + sids)
        # then notes
        if not notes:
            return
        nids = self.db.list(
            f"""select id from {self.username}.notes
            where id in {ids2str(nids)} and id not in (select nid from {self.username}.cards)"""
        )
        self.rem_notes(nids)

    def update_field_cache(self, nids):
        "Update field checksums and sort cache, after find&replace, etc."
        snids = ids2str(nids)

        r = []
        for (nid, mid, flds) in self.db.execute(f"select id, mid, flds from {self.username}.notes where id in {snids}"):
            fields = splitFields(flds)
            model = self.models.get(mid)
            if not model:
                # note points to invalid model
                continue
            r.append((stripHTMLMedia(fields[self.models.sort_idx(model)]), fieldChecksum(fields[0]), nid))
        # apply, relying on calling code to bump usn+mod
        self.db.executemany(f"update {self.username}.notes set sfld=%s, csum=%s where id=%s", r)

    def basic_check(self):
        "Basic integrity check for syncing. True if ok."
        # cards without notes
        # FIXME: this should be done with a foreign key!!!
        if self.db.scalar(
            f"""select 1 from {self.username}.cards
                where nid not in (select id from {self.username}.notes) limit 1"""
        ):
            return False
        # notes without cards or models
        # FIXME: this can be done with foreign keys when migrate to new collection tables structure
        if self.db.scalar(
            f"""select 1 from {self.username}.notes where id not in (select distinct nid from {self.username}.cards)
                                or mid not in {ids2str(self.models.ids())} limit 1"""
        ):
            return False
        # invalid ords
        for model in self.models.all():
            # ignore clozes
            if model["type"] != MODEL_STD:
                continue
            ids_string = ids2str([template["ord"] for template in model["tmpls"]])
            if self.db.scalar(
                f"""select 1 from {self.username}.cards where ord not in {ids_string} and nid in (
                    select id from {self.username}.notes where mid = %s) limit 1""",
                model["id"],
            ):
                return False
        return True

    # all new...
    def new_note(self, model):
        return Note(self, model)

    def set_note_review_in(self, note_id, review_in):
        # This method *should* allow one to change the due date. It DEFINITELY hasn't been properly
        # tested yet though!!!
        note = self.get_note(note_id=note_id)

        # see https://github.com/ankidroid/Anki-Android/wiki/Database-Structure
        # for more details of ids
        # type            integer not null,
        #   -- 0=new, 1=learning, 2=due, 3=filtered
        # queue           integer not null,
        #   -- -3=sched buried, -2=user buried, -1=suspended,
        #   -- 0=new, 1=learning, 2=due (as for type)
        #   -- 3=in learning, next rev in at least a day after the previous review
        # due             integer not null,
        #  -- Due is used differently for different card types:
        #  --   new: note id or random int
        #  --   due: integer day, relative to the collection's creation time
        #  --   learning: integer timestamp
        # odue            integer not null,
        #   -- original due: only used when the card is currently in filtered deck
        # odid            integer not null,
        #   -- original did: only used when the card is currently in filtered deck

        # TODO: we should also be able to add definitions from the UI

        # TODO: decide what to do about:
        # factor
        # ivl
        # filtered cards

        # FIXME: this may be way too dumb to actually work - I am assuming this will work not
        # only for cards that are for reviewing in the future but also new and learning cards
        # and that is just a blind guess! There may well be other fields/things that need to
        # be modified in addition to factor, ivl and filtered cards

        due = self.today() + review_in
        for card in note.cards():
            if card.type == 3:
                raise Exception("We don't know what to do with filtered cards yet, come back soon")

            card.type = 2
            card.queue = 2
            card.due = due
            card.flush()

        self.save()  # FIXME: Find a better way to unlock the db  - is this still necessary???

    def today(self):
        return int((time.time() - self.crt) // 86400)

    def create_note(self, note_json, deck_name, review_in=0):
        # models.by_name <- the model MUST exist
        model_name = note_json["model"]
        model = self.models.by_name(model_name)

        deck_id = self.decks.get_or_add(deck_name)

        note = self.new_note(model)
        note.model()["did"] = deck_id

        # print('my map', self.models.field_map(model))
        # print('my map', self.models.field_map(model).values())
        # raise Exception('here')
        # note_json["fields"]

        for field in self.models.field_map(model).values():
            note[field[1]["name"]] = note_json["fields"][field[0]]

        for tag in note_json["tags"]:
            if tag.strip():
                note.tags.append(tag)

        self.add_note(note)
        self.save()

        if review_in > 0:
            self.set_note_review_in(note.id, review_in)

        return note.id

    def add_note(self, note):
        "Add a note to the collection. Return number of new cards."
        # check we have card models available, then save
        cms = self.find_templates(note)
        if not cms:
            return 0
        note.flush()
        # deck conf governs which of these are used
        due = self.next_id("pos")
        # add cards
        ncards = 0
        for template in cms:
            self._new_card(note, template, due)
            ncards += 1
        return ncards

    def get_note(self, note_id):
        return Note(self, note_id=note_id)

    def find_templates(self, note):
        "Return (active), non-empty templates."
        model = note.model()
        avail = self.models.avail_ords(model, joinFields(note.fields))
        return self._templates_from_ordinals(model, avail)

    # def next_id(self, otype, inc=True):  we don't call with this it looks like so it's always True
    def next_id(self, otype):
        atype = "next" + otype.capitalize()
        an_id = self.conf.get(atype, 1)
        # if inc:
        #     self.conf[atype] = an_id + 1
        self.conf[atype] = an_id + 1

        return an_id

    @staticmethod
    def _templates_from_ordinals(model, avail):
        ok = []
        if model["type"] == MODEL_STD:
            for t in model["tmpls"]:
                if t["ord"] in avail:
                    ok.append(t)
        else:
            # cloze - generate temporary templates from first
            for ordi in avail:
                t = copy.copy(model["tmpls"][0])
                t["ord"] = ordi
                ok.append(t)
        return ok

    def _new_card(self, note, template, due, flush=True, did=None):
        "Create a new card."
        card = Card(self)
        card.nid = note.id
        card.ord = template["ord"]
        card.did = self.db.scalar(
            f"select did from {self.username}.cards where nid = %s and ord = %s", card.nid, card.ord
        )
        # Use template did (deck override) if valid, otherwise did in argument, otherwise model did
        if not card.did:
            if template["did"] and str(template["did"]) in self.decks.decks:
                card.did = template["did"]
            elif did:
                card.did = did
            else:
                card.did = note.model()["did"]
        # if invalid did, use default instead
        deck = self.decks.get(card.did)
        if deck["dyn"]:
            # must not be a filtered deck
            card.did = 1
        else:
            card.did = deck["id"]
        card.due = self._due_for_did(card.did, due)
        if flush:
            card.flush()
        return card

    def _due_for_did(self, did, due):
        conf = self.decks.conf_for_did(did)
        # in order due?
        if conf["new"]["order"] == NEW_CARDS_DUE:
            return due

        # random mode; seed with note ts so all cards of this note get the
        # same random number
        r = random.Random()
        r.seed(due)
        return r.randrange(1, max(due, 1000))

    def timestamp_for_table(self, table_name):
        "Return a non-conflicting timestamp for table."
        # Adapted from anki.utils.timestampID
        ts = intTime(1000)

        while self.db.scalar(f"select id from {self.username}.{table_name} where id = %s", ts):
            ts += 1
        return ts



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['djankiserv',
 'djankiserv.api',
 'djankiserv.assets.jsonfiles',
 'djankiserv.assets.sql.sqlite3',
 'djankiserv.migrations',
 'djankiserv.sync',
 'djankiserv.unki']

package_data = \
{'': ['*']}

install_requires = \
['django-k8s>=0.2.9,<0.3.0',
 'django>=3.1,<4.0',
 'djangorestframework-simplejwt>=4.4.0,<5.0.0',
 'djangorestframework>=3.11.1,<4.0.0',
 'gunicorn>=20.0.4,<21.0.0']

setup_kwargs = {
    'name': 'djankiserv',
    'version': '0.0.2',
    'description': 'Django-based synchronisation and API server for Anki',
    'long_description': None,
    'author': 'Anton Melser',
    'author_email': 'anton@melser.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

import json
import os
import pkgutil
from sqlite3 import dbapi2 as sqlite

from djankiserv.assets.sql import sqlite3  # noqa: F401 # pylint: disable=W0611
from djankiserv.unki import AnkiDataModelBase, intTime


class DB:  # this is only for the sqlite db
    def __init__(self, path):
        self._db = sqlite.connect(path, timeout=0)
        self._db.text_factory = self._textFactory
        self._path = path
        self.mod = False

    def execute(self, sql, *a):
        s = sql.strip().lower()
        # mark modified?
        for stmt in "insert", "update", "delete":
            if s.startswith(stmt):
                self.mod = True
                break

        return self._db.execute(sql, a)

    def executescript(self, sql):
        self.mod = True
        self._db.executescript(sql)

    def scalar(self, *a):
        res = self.execute(*a).fetchone()  # pylint: disable=E1120  # FIXME: how should I get rid of this disable?
        if res:
            return res[0]
        return None

    def close(self):
        self._db.text_factory = None
        self._db.close()

    def __enter__(self):
        self._db.execute("begin")
        return self

    def __exit__(self, exc_type, *args):
        self._db.close()

    # strip out invalid utf-8 when reading from db
    @staticmethod
    def _textFactory(data):
        return str(data, errors="ignore")


def sqlite3_for_download(path):
    path = os.path.abspath(path)
    if os.path.exists(path):
        raise Exception("This file must not exist!")

    base = os.path.basename(path)
    for c in ("/", ":", "\\"):
        assert c not in base

    db = DB(path)
    # db.isolation_level = None

    db.execute("pragma page_size = 4096")
    db.execute("pragma legacy_file_format = 0")
    db.execute("vacuum")

    db.executescript(pkgutil.get_data("djankiserv.assets.sql.sqlite3", "create_tables.sql").decode("utf-8"))
    db.executescript(
        pkgutil.get_data("djankiserv.assets.sql.sqlite3", "init_col_table.sql").decode("utf-8")
        % ({"v": AnkiDataModelBase.VERSION, "s": intTime(1000)})
    )

    c = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_collection_conf.json").decode("utf-8"))
    g = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_deck.json").decode("utf-8"))
    g["mod"] = intTime()
    gc = json.loads(pkgutil.get_data("djankiserv.assets.jsonfiles", "default_deck_conf.json").decode("utf-8"))

    db.execute(
        "update col set conf = ?, decks = ?, dconf = ?", json.dumps(c), json.dumps({"1": g}), json.dumps({"1": gc})
    )

    db.executescript(pkgutil.get_data("djankiserv.assets.sql.sqlite3", "add_indices.sql").decode("utf-8"))

    db.execute("analyze")
    db.execute("pragma temp_store = memory")
    db.execute("pragma cache_size = 10000")
    db.execute("pragma journal_mode = delete")  # this forces flush to disk when closing the file: MUST KEEP
    # db.execute("pragma journal_mode = wal")  # MUSTN'T use this

    # db.isolation_level = ""

    db.close()



# -*- coding: utf-8 -*-

import gzip
import io
import json
import re
import time
from abc import ABC, abstractmethod
from hashlib import sha1


def _decode_data(data, compression=0):
    if compression:
        with gzip.GzipFile(mode="rb", fileobj=io.BytesIO(data)) as gz:
            data = gz.read()
    try:
        data = json.loads(data.decode())
    except (ValueError, UnicodeDecodeError):
        data = {"data": data}

    return data


def get_data(request):
    try:
        compression = int(request.POST["c"])
    except KeyError:
        compression = 0

    try:
        data = _decode_data(request.FILES["data"].read(), compression)
        request.FILES["data"].seek(0)  # need to seek back to zero or it will appear empty after read()!
    except KeyError:
        data = {}
    return data


# from anki.utils
# FIXME: these have both been copied to multiple modules
def intTime(scale=1):
    "The time in integer seconds. Pass scale=1000 to get milliseconds."
    return int(time.time() * scale)


def ids2str(ids):
    """Given a list of integers, return a string '(int1,int2,...)'."""
    # This is "safe" for sql string concatenation due to casting to int before recasting to string,
    # meaning that python will raise an exception for anything but lists of numbers
    idstring = f"({','.join(str(int(i)) for i in ids)})"
    return idstring if len(idstring) > 2 else "(0)"


def splitFields(string):
    return string.split("\x1f")


def joinFields(field_list):
    return "\x1f".join(field_list)


def checksum(data):
    if isinstance(data, str):
        data = data.encode("utf-8")
    return sha1(data).hexdigest()


# deck schema & syncing vars
REM_CARD = 0
REM_NOTE = 1
# REM_DECK = 2  # declared and used only in unki.decks


# HTML
##############################################################################
reComment = re.compile("(?s)<!--.*?-->")
reStyle = re.compile("(?si)<style.*?>.*?</style>")
reScript = re.compile("(?si)<script.*?>.*?</script>")
reTag = re.compile("(?s)<.*?>")
reEnts = re.compile(r"&#?\w+;")
reMedia = re.compile("(?i)<img[^>]+src=[\"']?([^\"'>]+)[\"']?[^>]*>")


def entsToTxt(html):
    # entitydefs defines nbsp as \xa0 instead of a standard space, so we
    # replace it first
    html = html.replace("&nbsp;", " ")

    def fixup(m):
        text = m.group(0)
        if text[:2] == "&#":
            # character reference
            try:
                if text[:3] == "&#x":
                    return chr(int(text[4:-1], 16))
                return chr(int(text[2:-1]))
            except ValueError:
                pass
        else:
            # named entity
            try:
                # FIXME: name2codepoint doesn't exist anywhere!!!
                text = chr(name2codepoint[text[1:-1]])  # noqa: F821  # pylint: disable=E0602
            except KeyError:
                pass
        return text  # leave as is

    return reEnts.sub(fixup, html)


def stripHTML(s):
    s = reComment.sub("", s)
    s = reStyle.sub("", s)
    s = reScript.sub("", s)
    s = reTag.sub("", s)
    s = entsToTxt(s)
    return s


def stripHTMLMedia(s):
    "Strip HTML but keep media filenames"
    s = reMedia.sub(" \\1 ", s)
    return stripHTML(s)


def fieldChecksum(data):
    # 32 bit unsigned number from first 8 digits of sha1 hash
    return int(checksum(stripHTMLMedia(data).encode("utf-8"))[:8], 16)


AnkiDataModel = None  # This gets set as a module/global variable in the app init with the db connector (mysql or pgsql)


# This class is the source of truth for the Anki data model. It is used
# for various SQL generation purposes, such as creating schemas, query
# rewrites and getting all the tables for sqlite -> pg and pg -> sqlite
# import/export
class AnkiDataModelBase(ABC):
    VERSION = 11
    COLLECTION_PARENT_DB = "collection"
    MEDIA_PARENT_DB = "media"

    # is_pk gives a primary key constraint and an identity sequence, currently only supports single
    MODEL = {
        "notes": {
            "fields": [
                {"name": "id", "type": "bigint", "is_pk": True},
                {"name": "guid", "type": "text"},
                {"name": "mid", "type": "bigint"},
                {"name": "modified", "type": "bigint", "sqlite_name": "mod"},
                {"name": "usn", "type": "bigint"},
                {"name": "tags", "type": "text"},
                {"name": "flds", "type": "text"},
                {"name": "sfld", "type": "text"},  # WARNING! This is an `integer` in sqlite but contains text...
                # csum is actually an integer but must be text col because empty strings get inserted!
                # {"name": "csum", "type": "bigint", "nullable": True},
                {"name": "csum", "type": "text"},
                {"name": "flags", "type": "bigint"},
                {"name": "data", "type": "text"},
            ],
            "indexes": [
                {"name": "ix_notes_csum", "fields": [{"name": "csum", "type": "text"}]},
                {"name": "ix_notes_usn", "fields": [{"name": "usn", "type": "bigint"}]},
            ],
            "parent": COLLECTION_PARENT_DB,
        },
        "cards": {
            "fields": [
                {"name": "id", "type": "bigint", "is_pk": True},
                {"name": "nid", "type": "bigint"},
                {"name": "did", "type": "bigint"},
                {"name": "ord", "type": "bigint"},
                {"name": "modified", "type": "bigint", "sqlite_name": "mod"},
                {"name": "usn", "type": "bigint"},
                {"name": "type", "type": "bigint"},
                {"name": "queue", "type": "bigint"},
                {"name": "due", "type": "bigint"},
                {"name": "ivl", "type": "bigint"},
                {"name": "factor", "type": "bigint"},
                {"name": "reps", "type": "bigint"},
                {"name": "lapses", "type": "bigint"},
                {"name": "remaining", "type": "bigint", "sqlite_name": "left"},
                {"name": "odue", "type": "bigint"},
                {"name": "odid", "type": "bigint"},
                {"name": "flags", "type": "bigint"},
                {"name": "data", "type": "text"},
            ],
            "indexes": [
                {"name": "ix_cards_nid", "fields": [{"name": "nid", "type": "bigint"}]},
                {
                    "name": "ix_cards_sched",
                    "fields": [
                        {"name": "did", "type": "bigint"},
                        {"name": "queue", "type": "bigint"},
                        {"name": "due", "type": "bigint"},
                    ],
                },
                {"name": "ix_cards_usn", "fields": [{"name": "usn", "type": "bigint"}]},
            ],
            "parent": COLLECTION_PARENT_DB,
        },
        "col": {
            "fields": [
                {"name": "id", "type": "bigint", "is_pk": True},
                {"name": "crt", "type": "bigint"},
                {"name": "modified", "type": "bigint", "sqlite_name": "mod"},
                {"name": "scm", "type": "bigint"},
                {"name": "ver", "type": "bigint"},
                {"name": "dty", "type": "bigint"},
                {"name": "usn", "type": "bigint"},
                {"name": "ls", "type": "bigint"},
                {"name": "conf", "type": "text"},
                {"name": "models", "type": "text"},
                {"name": "decks", "type": "text"},
                {"name": "dconf", "type": "text"},
                {"name": "tags", "type": "text"},
            ],
            "indexes": [],
            "parent": COLLECTION_PARENT_DB,
            "initsql": (
                "insert into {schema_name}.col values " f"(1, 0, 0, 1, {VERSION}, 0, 0, 0, '', '{{}}', '', '', '{{}}');"
            ),
        },
        "graves": {
            "fields": [
                {"name": "usn", "type": "bigint"},
                {"name": "oid", "type": "bigint"},
                {"name": "type", "type": "bigint"},
            ],
            "indexes": [],
            "parent": COLLECTION_PARENT_DB,
        },
        "revlog": {
            "fields": [
                {"name": "id", "type": "bigint", "is_pk": True},
                {"name": "cid", "type": "bigint"},
                {"name": "usn", "type": "bigint"},
                {"name": "ease", "type": "bigint"},
                {"name": "ivl", "type": "bigint"},
                {"name": "lastivl", "type": "bigint"},
                {"name": "factor", "type": "bigint"},
                {"name": "rtime", "type": "bigint", "sqlite_name": "time"},
                {"name": "type", "type": "bigint"},
            ],
            "indexes": [
                {"name": "ix_revlog_cid", "fields": [{"name": "cid", "type": "bigint"}]},
                {"name": "ix_revlog_usn", "fields": [{"name": "usn", "type": "bigint"}]},
            ],
            "parent": COLLECTION_PARENT_DB,
        },
        "media": {
            "fields": [
                {"name": "fname", "type": "text", "is_pk": True},
                {"name": "usn", "type": "bigint"},
                {"name": "csum", "type": "text", "nullable": True},
            ],
            "indexes": [{"name": "ix_media_usn", "fields": [{"name": "usn", "type": "bigint"}]}],
            "parent": MEDIA_PARENT_DB,
        },
        "meta": {
            "fields": [{"name": "dirmod", "type": "bigint"}, {"name": "lastusn", "type": "bigint"}],
            "indexes": [],
            "parent": MEDIA_PARENT_DB,
            "initsql": "insert into {schema_name}.meta values (0, 0);",
        },
    }

    @staticmethod
    @abstractmethod
    def generate_schema_sql_list(schema_name):
        pass

    @staticmethod
    @abstractmethod
    def insert_on_conflict_update(schema_name, table_name):
        pass

    @staticmethod
    @abstractmethod
    def insert_on_conflict_nothing(schema_name, table_name):
        pass

    @staticmethod
    @abstractmethod
    def replace_schema(cur, to_replace_name, replace_with_name):
        pass
